from pulsar import send
from pulsar.apps import http
from jirachi.io.abstract import JirachiMonitor

__all__ = ['RequestMonitor']


class RequestMonitor(JirachiMonitor):
    name = 'request_worker'

    async def monitor_start(self, monitor, exec=None):
        monitor.sessions = http.HttpClient()

    async def monitor_stopping(self, monitor, exec=None):
        if hasattr(monitor, 'sessions'):
            monitor.sessions

    async def _request(actor, url, method, **args):
        async with actor.sessions as sessions:
            res = await getattr(sessions, method.lower())(url, **args)
        return res

    @classmethod
    async def request(cls, url, method='GET', *args, **kwargs):
        monitor = await cls.get_monitor()
        return await send(monitor, 'run', cls._request, url, method, *args, **kwargs)



from pulsar import send
from motor.motor_asyncio import AsyncIOMotorClient
from jirachi.io.abstract import MongoDBMonitor

__all__ = ['MongoDBMonitor']


class MongoDBMonitor(MongoDBMonitor):
    name = "mongodb"

    async def monitor_start(self, monitor, exec=None):
        monitor.client = await AsyncIOMotorClient(self.cfg.mongoconf)

    async def _insert_one(actor, db, collect, data):
        return await actor[db][collect].insert_one(data)

    async def _find_one(actor, db, collect, data):
        return await actor[db][collect].find_one(data)

    @classmethod
    async def insert_one(cls, db: str, collect: str, data: dict) -> dict:
        monitor = await cls.get_monitor()
        return await send(monitor, 'run', cls._insert_one, db, collect, data)

    @classmethod
    async def find_one(cls, db: str, collect: str, data: dict) -> dict:
        monitor = await cls.get_monitor()
        return await send(monitor, 'run', cls._find_one, db, collect, data)



from typing import Iterable
import asyncpg
from asyncpg import Record
from pulsar import send
from jirachi.io.abstract import JirachiMonitor


__all__ = ['PostgresMonitor']


class PostgresMonitor(JirachiMonitor):
    name = "postgres"

    async def connect(self, *args, **kwargs):
        return await asyncpg.connect(**self.cfg.pgconfigs)

    @staticmethod
    async def _execute(actor, sql) -> str:
        return await actor.conn.execute(sql)

    @staticmethod
    async def _fetch(actor, sql) -> dict:
        res = await actor.conn.fetchrow(sql)
        return dict(res)

    @staticmethod
    async def _transaction(actor, sqls) -> str:
        async with actor.conn.transaction():
            res = [await actor.conn.execute(s) for s in sqls]
        return res

    async def monitor_start(self, monitor, exec=None):
        monitor.conn = await asyncpg.connect(**self.cfg.pgconf)

    async def monitor_stopping(self, monitor, exec=None):
        if hasattr(monitor, 'conn'):
            monitor.conn.terminate()

    async def worker_start(self, worker, exc=None):
        if not worker.is_arbiter or worker.is_monitor:
            worker.conn = await asyncpg.connect(**self.cfg.pgconf)

    async def worker_stopping(self, worker, exc=None):
        if not worker.is_arbiter or worker.is_monitor:
            if hasattr(worker, 'conn'):
                worker.conn.terminate()

    @classmethod
    async def execute(cls, sql: str) -> str:
        monitor = await cls.get_monitor()
        return await send(monitor, 'run', cls._execute, sql)

    @classmethod
    async def fetch(cls, sql: str) -> Record:
        monitor = await cls.get_monitor()
        return await send(monitor, 'run', cls._fetch, sql)

    @classmethod
    async def transaction(cls, sqls: Iterable[str]) -> str:
        monitor = await cls.get_monitor()
        return await send(monitor, 'run', cls._transaction, sqls)



from pulsar import get_application, get_actor, send
from pulsar.apps import Application

__all__ = ['JirachiMonitor', 'JirachiMonitorNotFound']


class JirachiMonitorNotFound(Exception):
    pass


class JirachiMonitor(Application):

    @classmethod
    async def get_arbiter(cls):
        arbiter = get_actor().get_actor('arbiter')
        return arbiter

    @classmethod
    async def get_monitor(cls):
        async def get_monitor_via_arbiter():
            arbiter = get_actor().get_actor('arbiter')
            monitor_name = next(
                (m for m in arbiter.monitors if cls.name in m), None)
            monitor = await get_application(monitor_name)
            return monitor
        name = cls.cfg.name or cls.name
        monitor = get_actor().get_actor(name)
        return monitor or await get_monitor_via_arbiter()

    @classmethod
    async def kill(cls):
        monitor = await cls.get_monitor()
        return await send('arbiter', 'kill_actor', monitor.name)



from typing import Any
from collections import defaultdict

__all__ = ['configdict']


def configdict(cases: dict, other: Any) -> defaultdict:
    assert other in cases.keys(), 'cases should include default cases'
    return defaultdict(lambda: other)(cases)



from pulsar.apps import MultiApp
from pulsar.apps.wsgi.handlers import WsgiHandler
from pulsar.apps.wsgi import WSGIServer
from jirachi.core import wsgi
from jirachi.io import RequestMonitor
from jirachi.io import SchedulerMonitor

__all__ = ['wsgi', 'ComposedApp', 'ComposedIO',
           'RequestMonitor', 'SchedulerMonitor']


__version__ = '0.1.2.5'


class ComposedIO(MultiApp):
    name = 'io'

    def build(self):
        yield self.new_app(RequestMonitor, worker=10)
        yield self.new_app(SchedulerMonitor, worker=10)


class ComposedApp(MultiApp):
    name = 'jirachi'

    def build(self):
        yield self.new_app(WSGIServer, callable=WsgiHandler((wsgi, )),)
        yield self.new_app(ComposedIO, worker=10)



import unittest
from pulsar import send
from jirachi.io.requests import RequestMonitor


class TestRequest(unittest.TestCase):
    @classmethod
    async def setUpClass(cls):
        cls.monitor = RequestMonitor(workers=10)
        await send('arbiter', 'run', cls.monitor)

    async def testGet(self):
        res = await RequestMonitor.request(url='http://www.baidu.com', method='GET')
        self.assertTrue(res)

    @classmethod
    def tearDownClass(cls):
        return send('arbiter', 'kill_actor', 'request_worker')



from functools import partial
import unittest
from pulsar import send
from pulsar import Config
from jirachi.io.scheduler import SchedulerMonitor
import asyncio

monitor = SchedulerMonitor(workers=1, name='scheduler_worker')

res = []


@partial(monitor.task, rule=lambda t: int(t) % 1 == 0)
def task():
    res.append('a')


class TestScheduler(unittest.TestCase):
    @classmethod
    async def setUpClass(cls):
        await send('arbiter', 'run', monitor)

    @classmethod
    def tearDownClass(cls):
        return send('arbiter', 'kill_actor', 'scheduler_worker')

    async def testMeta(self):
        await asyncio.sleep(5)
        self.assertTrue(res)



# -*- eval: (venv-workon "jirachi"); -*-

import unittest
from pulsar import send, get_actor
from pulsar import Config
from jirachi.io.remote import RemoteMonitorWSGI
from pulsar.apps.http import HttpClient


class TestRemoteMonitor(unittest.TestCase):

    @classmethod
    async def setUpClass(cls):
        cls.monitor = RemoteMonitorWSGI(workers=5, bind="127.0.0.1:8964")
        cls.app_cfg = await send('arbiter', 'run', cls.monitor)
        cls.client = HttpClient()
        cls.uri = 'http://{0}:{1}'.format(*cls.app_cfg.addresses[0])

    async def test_basic(self):
        url = self.uri + '/remote/remote_monitor/event/test/'
        resp = await self.client.post(url)
        self.assertEqual(resp.status_code, 200)
        m = await self.monitor.get_monitor()
        self.assertEqual(m.name, 'remote_monitor')
        self.assertTrue(get_actor().get_actor(m.name))

    @classmethod
    def tearDownClass(cls):
        return send('arbiter', 'kill_actor', 'remote_monitor')



import unittest
from pulsar import send, get_application, get_actor
from jirachi import ComposedApp


class TestComposedApp(unittest.TestCase):
    @classmethod
    async def setUpClass(cls):
        cls.monitor = ComposedApp()
        cls.app_cfg = await send('arbiter', 'run', cls.monitor)

    async def test_get_monitor(self):
        app = await get_application('jirachi')
        print(self.monitor.apps()[1].apps())
        arbiter = get_actor().get_actor('arbiter')
        self.assertTrue(app)
        self.assertTrue(arbiter)

    @classmethod
    async def tearDownClass(cls):
        for app in cls.monitor.apps():
            await send('arbiter', 'kill_actor', app.name)
        return



import time
from functools import wraps
from jirachi.io.abstract import JirachiMonitor

__all__ = ['SchedulerMonitor']


class SchedulerMonitor(JirachiMonitor):
    name = 'scheduler_worker'
    _tasks = []

    def task(self, fn, rule=lambda t: int(t) % 1 == 0):
        self._tasks.append([rule, fn])

        @wraps(fn)
        def _(*args, **kwargs):
            return fn(*args, **kwargs)
        return _

    def monitor_task(self, monitor):
        [t[1]() for t in self._tasks if t[0](time.time())]



from hashlib import sha1
import time
import json
from pulsar import get_actor, Future
from pulsar.apps.wsgi import WSGIServer, WsgiResponse, WsgiHandler
import aio_etcd as etcd
from jirachi.io.abstract import JirachiMonitor, JirachiMonitorNotFound
from pulsar.apps.wsgi import Router

__all__ = ['RemoteMonitorWSGI']

blueprint = Router('/')


@blueprint.router('/remote/<string:monitor>/event/<string:event>/', methods=['post'])
async def remote_wsgi(request):
    hash_code = sha1(str(time.time()).encode()).hexdigest()
    monitor_name = request.urlargs['monitor']
    event_name = request.urlargs['event']
    data = request.body_data

    actor = get_actor()
    monitor = actor.get_actor(monitor_name)
    if monitor:
        monitor.fire_event(event_name, data)
    elif not (actor.is_arbiter() or actor.is_monitor() and actor.monitor == monitor):
        actor.monitor.fire_event(event_name, msg=data)
    else:
        raise JirachiMonitorNotFound('Cant found Monitor %s' % monitor)
    return WsgiResponse(200, json.dumps({
        'successed': True,
        'token': hash_code
    }))


class RemoteMonitorWSGI(WSGIServer, JirachiMonitor):
    name = 'remote_monitor'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cfg.callable = WsgiHandler((blueprint, ))
        if not hasattr(self.cfg, 'blacklist'):
            self.cfg.blacklist = []

    @staticmethod
    async def singlecast(msg, actor_name):
        actor = get_actor()
        if not actor.name == actor_name:
            actor = actor.get_actor(actor_name)
        if not actor and actor.monitor.name == actor.name:
            actor = actor.monitor
        actor.fire_event('singlecast', msg)
        actor.future = Future()
        return actor.future

    @staticmethod
    def event_test(msg):
        print('test event %s' % msg)

    async def monitor_start(self, monitor, exec=None):
        monitor.bind_event('test', self.event_test)
        if not hasattr(self.cfg, 'etcdconf'):
            monitor.etcd = etcd.Client()
        else:
            monitor.etcd = etcd.Client(**self.cfg.etcdconf)
        await super().monitor_start(monitor)

    async def worker_start(self, worker, *args, **kwargs):
        worker.bind_event('test', self.event_test)

    async def search_remote(self):
        pass

    async def sync_remote(self):
        pass
